;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; File:         db-search.l
; Description:  Frolic support for reasoning about db-objects
; Author:       Jed Krohnfeldt, University of Utah
; Created:      11-Aug-86 (structure-sharing version December 1987)
; Package:      FROLIC
;
; (c) Copyright 1986, 1987, University of Utah, all rights reserved.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package 'frolic)

(export '(cr rm as er ma m-a m-r m-e sh
             create remove assert match match-assert
             match-remove match-erase))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; db pattern processing
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; A db-pattern looks like #(keyword _name [class] (s1 _v1) (s2 _v2)...),
;; where [] indicate class is optional.

(defmacro db-pattern? (object)
  `(and (vectorp ,object) (not (stringp ,object))))

(defmacro db-pat? (object)
  `(and (molecule-p ,object) (db-pattern? (mol-skel ,object))))

;; Grab the keyword from a db-pattern.

(defmacro db-pattern-keyword (db-pattern)
  `(aref ,db-pattern 0))

;; Grab the db object name from a db-pattern.

(defmacro db-pattern-name (db-pattern)
  `(aref ,db-pattern 1))

;; The third argument in a db-pattern is an optional class (type) designation.

(defmacro db-pattern-optional-class (db-pattern)
  `(and (> (length ,db-pattern) 2) (aref ,db-pattern 2)))

;; If the third argument in a db-pattern is a symbol, it is a class (type)
;; specification for the pattern name.

(defmacro class-provided? (db-pattern)
  `(let ((class (db-pattern-optional-class ,db-pattern)))
     (and class (symbolp class))))

;; Distinguish between _var and var.

(defmacro pattern-name (name)
  `(if (var-name? ,name)
     (intern (subseq (symbol-name ,name) 1 nil))
     ,name))

;; Get the class for a db-pattern - if the optional class parameter is
;; specified, use it; otherwise, extract the db-pattern class from the name.

(defmacro db-pattern-class (db-pattern)
  `(let ((class (if (class-provided? ,db-pattern) 
		  (db-pattern-optional-class ,db-pattern)
		  (pattern-name (db-pattern-name ,db-pattern)))))
     (if (var? class) (pattern-name (var-name class)) class)))

;; Get a text version of the db-pattern name.

(defmacro db-text-name (db-name)
  `(if (symbolp ,db-name) 
     ,db-name
     (error "Unknown class (~S)~%" ,db-name)))

;; Grab all of the db-pattern's slots.

(defmacro db-pattern-slots (db-pattern)
  `(if (class-provided? ,db-pattern)
     (cdddr (coerce ,db-pattern 'list))
     (cddr (coerce ,db-pattern 'list))))

;; Get all of the db-objects that will match a given functor.

(defun get-db-objects (db-pattern)
  (let ((class (db-pattern-class db-pattern)))
    (if (db-class? class)
      (db-instances class)
      (list class))))

;; Functor indexing - handles db-patterns in rules
;; (redefinition of version in prolog.l to handle db objects).

(defun index-rule (skeleton num-vars)
  (let ((func (principal-functor skeleton)))
    (cond ((db-pattern? func)
	   (add-rule-to-index
 	     skeleton (db-text-name (db-pattern-class func)) num-vars))
	  ((symbolp func)
	   (add-rule-to-index skeleton func num-vars)))))

;; Coerce db-pattern to list for unification.

(defmacro db-pattern-to-list (db-pattern)
  `(coerce ,db-pattern 'list))

;; Parse the argument list of a db-pattern message.
;; The argument list looks like:
;;   (:in arg1 arg2 arg3 :out pat1 pat2 pat3), or 
;;   (pat1 pat2 pat3) which is equivalent to (:out pat1 pat2 pat3).
;; The :in are values to be sent to the object.
;; The :out are patterns to be unified with the object's return values.
; 
; (defun parse-db-pattern (args)
;   (let ((out-seen nil)
; 	(in-seen nil)
;         (msg-args nil)
; 	(out-pats nil))
;     (dolist (arg args)
;       (if out-seen
; 	(push arg out-pats)
;         (case arg
; 	  (:in (setf in-seen t))
; 	  (:out (setf out-seen t))
; 	  (otherwise
; 	    (if in-seen
; 	      (push arg msg-args)
; 	      (push arg out-pats))))))
;     (values (nreverse msg-args) (nreverse out-pats))))
; 

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; db pattern operations (create, remove, assert, erase)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Create a new db object.

(defun create-db-object (db-molecule goals level back)
  (let* ((db-pattern (mol-skel db-molecule))
	 (env (mol-env db-molecule))
	 (class (db-pattern-class db-pattern))
	 (new-slots nil)
	 (db-mol nil))
    (dolist (query (db-pattern-slots db-pattern))
      (let ((slot (car query))
	    (value (cadr query)))
	(if (var? slot)
	  (setf slot (expand-logical-vars slot env)))
	(if (var? value)
	  (setf value (expand-logical-vars value env)))
	(push `(,slot ',value) new-slots)))
    ; create a new db object
    (setf db-mol (make-molecule (db-new-instance class :init new-slots) env))
    (match-db-class db-pattern env (mol-skel db-mol) (mol-env db-mol))
    (pl-search (rest goals) level back)))

;; Remove a db object.

(defun remove-db-object (db-molecule goals level back)
  (let* ((db-pattern (mol-skel db-molecule))
	 (env (mol-env db-molecule))
	 (db-object (db-pattern-name db-pattern)))
    (if (var? db-object)
      (setf db-object (expand-logical-vars db-object env)))
    (if (db-instance? db-object)
      (progn
	(db-kill db-object)
	(succeed-continue db-pattern
			  (rest goals)
			  `(item ,db-object)
			  level
			  back))
      *impossible*)))

;; Assert slots in a db object.

(defun assert-db-object-slots (db-molecule goals level back)
  (let* ((db-pattern (mol-skel db-molecule))
	 (env (mol-env db-molecule))
	 (db-object (db-pattern-name db-pattern)))
    (if (var? db-object)
      (setf db-object (expand-logical-vars db-object env)))
    (if (db-instance? db-object)
      (progn
	(dolist (query (db-pattern-slots db-pattern) env)
	  (let ((slot (car query))
	        (value (cadr query)))
	    ; get binding for slot if it is a logical-variable
	    (if (var? slot)
             (setf slot (expand-logical-vars slot env)))
	    ; assert slot with value
	    (db-assert db-object slot (expand-logical-vars value env))))
	(succeed-continue db-pattern
			  (rest goals)
			  `(item ,db-object)
			  level
			  back))
      *impossible*)))
      
;; Erase slots in a db object.

(defun erase-db-object-slots (db-molecule goals level back)
  (let* ((db-pattern (mol-skel db-molecule))
	 (env (mol-env db-molecule))
	 (db-object (db-pattern-name db-pattern)))
    (if (var? db-object)
      (setf db-object (expand-logical-vars db-object env)))
    (if (db-instance? db-object)
      (progn
	(dolist (query (db-pattern-slots db-pattern) env)
	  (let ((slot (car query)))
	    ; get binding for slot if it is a logical-variable
	    (if (var? slot)
              (setf slot (expand-logical-vars slot env)))
	    ; remove slot from db-object
	    (db-remove db-object slot)))
	(succeed-continue db-pattern
			  (rest goals)
			  `(item ,db-object)
			  level
			  back))
      *impossible*)))
      
;; Attempt to match db-pattern against the db database.

(defun match-db-object-only (db-molecule goals level back)
  (let ((dbs (get-db-objects (mol-skel db-molecule))))
    (search-db goals dbs level back)))

;; Attempt to match db-pattern against the db database - if that fails, try
;; the prolog rulebase.

(defun match-db-object (db-molecule goals level back)
  (let* ((db-pattern (mol-skel db-molecule))
	 (dbs (get-db-objects db-pattern)))
    (search-db goals dbs level back)))
#|
    ; if no match against the database, and the db-pattern
    ; is a match pattern, try the rulebase
    (if (impossible? env)
      (let ((prolog-rules
	     (get-rule-molecules
	      (db-text-name (db-pattern-class db-pattern)))))
	(if (null prolog-rules)
	  *impossible*
	  (search-rules goals prolog-rules level back)))
      env)))
|#
;; Handlers for db pattern keywords.

(defun run-db-pattern-handler (db-molecule goals level back)
  (let ((keyword (db-pattern-keyword (mol-skel db-molecule))))
    (case keyword
      ((create cr) (create-db-object db-molecule goals level back))
      ((remove rm) (remove-db-object db-molecule goals level back))
      ((assert as) (assert-db-object-slots db-molecule goals level back))
      ((erase er) (erase-db-object-slots db-molecule goals level back))
      ((match ma) (match-db-object db-molecule goals level back))
      ((match-assert m-a match-remove m-r match-erase m-e show sh)
       (match-db-object-only db-molecule goals level back))
      (otherwise (error "Illegal db pattern ~S~%" db-molecule)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; unification and search
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Return environment in which x and y are unified.
;; (redefinition of unify2 in prolog.l to handle db objects)

(defun unify2 (x y)
  (cond ((var? x) (pl-bind x *x-env* y *y-env*))
	; bind variable if distinct
	((var? y) (pl-bind y *y-env* x *x-env*))
	; one term is a db object, the other is a db pattern
	((and (db-instance? x) (db-pattern? y))
	 (match-db y *y-env* x *x-env*))
	((and (db-pattern? x) (db-inst? y))
	 (match-db x *x-env* y *y-env*))
	; both terms db-patterns
	((and (db-pattern? x) (db-pattern? y))
	 (unify1 (db-pattern-to-list x) (db-pattern-to-list y)))
	; unify atoms
	((atom x) (if (equalp x y) t *impossible*))
	((atom y) *impossible*)
        ; both terms complex
	((let ((x-env *x-env*)
	       (y-env *y-env*)
	       (ans (unify1 (car x) (car y))))
	   (if (impossible? ans)
	     *impossible*
	     (progn
	       (setf *x-env* x-env)
	       (setf *y-env* y-env)
	       (unify1 (cdr x) (cdr y))))))))

;; Database version of database search for prolog.
;; (redefinition of version in prolog.l to handle db interface)

(defun pl-search (goals level back)
  (if (null goals)
    (succeed back)
    (let ((goal (first goals)))
      ; if goal is a db pattern, run db pattern handler
      (cond ((db-pat? goal)
	     (run-db-pattern-handler goal goals level back))
	    (t
	     (search-rules goals
	       (get-rule-molecules (goal-functor (first goals)))
	       level back))))))

;; Match goal(s) against database.

(defun search-db (goals rem-db level back)
  (do ((goal (first goals))
       (rules rem-db (rest rules))
       (old-trail *trail*)
       (trail *trail*))
      ((null rules) (fail-continue goal back))
    (setf trail *trail*)
    (if (impossible?
	 (match-db (mol-skel goal) (mol-env goal) (first rules) nil))
      (untrail trail)
      (return
       (succeed-continue goal
			 (rest goals)
			 (first rules)
			 level
			 (make-cont goals	    ; goals
				    (rest rules)    ; rules
				    level	    ; level
				    back	    ; back
				    old-trail	    ; trail
				    nil))))))	    ; trace

;; Continue searching with continuation - backtrack to a choice point and
;; continue executing.
;; (redefinition of continue-on in prolog.l to handle db objects)

(defun continue-on (cont)
  (if (null cont)
    *impossible*
    (if (null (cont-goals cont))
      (continue-on (cont-back cont))
      (progn
	; wrap trail back to last choice point, undoing bindings
	(untrail (cont-trail cont))
	; resume search
	(cond ((db-pat? (first (cont-goals cont)))
	       (search-db (cont-goals cont) (cont-rules cont)
			  (cont-level cont) (cont-back cont)))
	      (t 
	       (search-rules (cont-goals cont) (cont-rules cont)
			     (cont-level cont) (cont-back cont))))))))

;; Simple version of match-db with no message passing or multiple values.
;; Try to match pattern to db-object, adding logical bindings to env.

(defun match-db (pat-obj pat-env db-obj db-env)
  (let ((return-env (match-db-class pat-obj pat-env db-obj db-env)))
    (if (not (impossible? return-env ))
      (dolist (query (db-pattern-slots pat-obj))
	(let ((slot-name (car query))
	      (pat-val (cadr query))
	      (db-val nil))
	  ; get the actual value of the slot
	  (setf db-val (db-ask db-obj slot-name))
	  ; compare the return value against the output pattern,
	  ; adding logical variables to the environment if necessary
	  (setf return-env (unify pat-val pat-env db-val db-env))
	  (if (impossible? return-env)
	    (return return-env)))))
    (if (not (impossible? return-env))
      ; if we get here, the pattern matched the db-obj completely
      ; now perform the side effect operation if any is required
      (case (db-pattern-keyword pat-obj)
	((match-remove m-r) (db-kill db-obj))
	((show sh) (db-show db-obj))
	((match-erase m-e) (dolist (query (db-pattern-slots pat-obj))
		       (db-remove db-obj (car query))))
	((match-assert m-a) (dolist (query (db-pattern-slots pat-obj))
			(db-assert db-obj (car query)
                                   (expand-logical-vars
                                    (caddr query)
                                    pat-env))))))
    return-env))

;; Match the class name of the pattern against the class of the db-object,
;; augmenting the environment with logical bindings if successful.

(defun match-db-class (pat-obj pat-env db-obj db-env)
  (let ((pat-class (db-pattern-class pat-obj))
	(pat-name (db-pattern-name pat-obj))
	(db-class (db-class db-obj)))
    ; class of pattern is same as object or as class of object
    (if (or (equal pat-class db-obj) (equal pat-class db-class))
      (if (var? pat-name)
	(unify pat-name pat-env db-obj db-env)
	t)
      *impossible*)))

;; Redefined from prolog.l to handle db-patterns.

(defun pl-assert (rule)
  (let ((head (head rule))
	(body (body rule)))
    ; if head is a db-pattern with side effects, call pl-solve-all
    ; since it handles db-patterns correctly
    (if (and (db-pattern? head)
	     (member (db-pattern-keyword head)
		     '(create cr remove rm erase er assert as)))
      (if (null body)
	(pl-solve-all (list head))
	(error "Ill-formed db interface clause ~S~%" rule))
      ; ordinary prolog rule
      (let ((env (list nil)))
	(setf *num-slots* -1)
	(index-rule (calcify rule env) (1+ *num-slots*)))))
  rule)

;; End of file.
