;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; File:         frob-db.l
; Description:  db interface for frobs
; Author:       Jed Krohnfeldt, University of Utah
; Created:      11-Aug-86
; Language:     Lisp
; Package:      FROLIC
; Status:       Experimental (Do Not Distribute)
;
; (c) Copyright 1986, University of Utah, all rights reserved.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; This file implements an interface between the db interface and the frob
; (frame + object) data structures implemented at Utah.  The db interface
; is a standard interface to slot-filler data structures, and is documented
; elsewhere.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package 'frolic)

;(export '(db-new-instance db-assert db-ask db-remove db-kill
;	    db-class db-class? db-classes db-instance? db-instances
;	    db-find db-show db-reset db-pat-instance))

;; Return the frob-class for the class name given.

(defmacro name-class (name)
  `(frobs::get-class-frob ,name))

;; Return the name of a frob-class.

(defmacro class-name (class)
  `(frobs::class-name ,class))

;; Return the frob-instance for the instance name given.

(defmacro name-instance (name)
  `(frobs::get-frob ,name))

;; Return the name of a frob-instance.

(defmacro instance-name (instance)
  `(frobs::instance-name ,instance))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; db interface functions
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Define a new db object (instance of a class).

(defun db-new-instance (class &key (name nil) (init nil))
  (let ((the-class (name-class class)))
    (if (not the-class)
      (error "Unknown frob class" class)
      (progn
	; create the generic instance of the frob
	(instance-name (frobs::new-instance* the-class
					     :init `((nil ,@init))))))))

;; Slot update function.

(defun db-assert (frob slot value)
  (frobs::assert-val (name-instance frob) slot value))

;; Slot access function.

(defun db-ask (frob slot)  
  (frobs::ask (name-instance frob) slot))

;; Remove a value from a slot.

(defun db-remove (frob slot &optional (value t value?))
  (if value?
    (frobs::remove-val (name-instance frob) slot value)
    (frobs::erase (name-instance frob) slot)))

;; Destroy a db object.

(defun db-kill (frob) 
  (frobs::kill-frob (name-instance frob)))

;; Return the class name of a db object.

(defun db-class (instance)
  (frobs::frob-type (name-instance instance)))

;; Return t if name is a db class name.

(defun db-class? (name)  
  (frobs::class-frob? (name-class name)))

;; Kill a class db object.

(defun db-kill-class (class)
  (frobs::kill-class-frob (name-class class)))

;; Return names of all db classes.

(defun db-classes ()
  (frobs::all-class-frobs))

;; Returns t if thing is a db object.

(defun db-instance? (thing)  
  (frobs::instance-frob? (name-instance thing)))

;; Return all instances of a db class.

(defun db-instances (class-name)
  (let ((class (name-class class-name)))
    (if class
      (mapcar #'frobs::instance-name (frobs::instances class)))))

;; Limited solve, pattern matching.

(defun db-find (class slot value)
  (declare (ignore class slot value))
  nil)

;; Prettyprint a db object.

(defun db-show (instance)
  (frobs::show (name-instance instance)))

;; Send a message to a db object.

(defun db-send (obj msg &rest args)
  (funcall msg `(,(name-instance obj) ,@args)))

;; Reset all of the db objects.

(defun db-reset ()
  (frobs::reset-system))

;; End of file.
