;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; File:         frob-search.l
; Description:  prolog-style unification for frobs
; Author:       Jed Krohnfeldt, University of Utah
; Created:      5-Aug-86
; Modified:     6-Aug-86 09:34:10 (Jed Krohnfeldt)
; Language:     Lisp
; Package:      USER
; Status:       Experimental (Do Not Distribute)
;
; (c) Copyright 1986, University of Utah, all rights reserved.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; this file should be loaded after prolog.l is loaded
; it redefines the pl-search and unify routines to handle reasoning about frobs
; (frobs in frobs.b courtesy of Eric Muehle)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmacro frob-pattern? (object)
  `(arrayp ,object))

;; frob-name - grab the frob name for a frob query

(defmacro frob-pat-name (frob-pattern)
  `(aref ,frob-pattern 0))

;; frob-text-name - get a text version of the frob pattern name

(defmacro frob-text-name (frob-name)
  `(cond ((var-name? ,frob-name) (frob-class ,frob-name))
	 ((or (frob-p ,frob-name) (class-frob-p ,frob-name))
	  (frob-name ,frob-name))
	 ((symbolp ,frob-name) ,frob-name)))

;; frob-slots - grab all of the frob-pattern's slots

(defmacro frob-slots (frob-pattern)
  `(cdr (coerce ,frob-pattern 'list)))

;; frob-class - get the class for a logical variable frob pattern

(defmacro frob-class (frob-name)
  `(intern (subseq (symbol-name ,frob-name) 1 nil)))

;; get-class-frobs - get all of the frobs that are children of class

(defmacro get-class-frobs (class)
  `(let ((ret-val nil))
     (dolist (x (class-frob-children (gethash ,class *class-hash*)) ret-val)
	(push (gethash x *hash*) ret-val))))

;; get-frobs - get all of the frobs that will match a given functor

(defmacro get-frobs (functor)
  `(let ((frob nil))
     (cond ((var-name? ,functor)
	    (get-class-frobs (frob-class ,functor)))
	   ((or (frob-p ,functor) (class-frob-p ,functor))
	    (list ,functor))
	   ((symbolp ,functor)
	    (setf frob (gethash ,functor *hash*))
	    (if frob (list frob))))))

;; functor indexing - handles frob rules
;; (redefinition of version in search.cl to handle frobs)

(defun index-rule (rule)
  (let ((func (principal-functor rule)))
    (cond ((frob-pattern? func)
	   (add-rule-to-index rule (frob-text-name (frob-pat-name func))))
	  ((symbolp func)
	   (add-rule-to-index rule func)))))

;; frob-to-list - coerce frob pattern to list for unification

(defmacro frob-to-list (frob)
  `(cons (tag-part ,frob)
	 (coerce (term-part ,frob) 'list)))

;; unify1 - returns environment in which x and y are unified
;; (redefinition of version in search.cl to handle frobs)

(defun unify1 (env x y)
  (cond ((var? x) (if (eqvars x y) env (bind x y env)))
	; bind variable if distinct
	((var? y) (bind y x env)) 
	; both terms frob patterns
	((and (frob-pattern? (term-part x)) (frob-pattern? (term-part y)))
	 (unify-complex env (frob-to-list x) (frob-to-list y)))
	((atom? x)  (if (eqatoms x y) env *impossible*))
	; atoms must be eq in order to unify
	((atom? y)  *impossible*)  
	; both terms complex
	(t (unify-complex env x y)))) 

;; pl-search - frob version (original in search.cl)
;; (redefinition of version in search.cl to handle frobs)

(defun pl-search (env goals level back)
  (if (null goals)
    (succeed env back)
    (let ((goal (first goals))
	  (frobs nil))
      ; if goal is a frob patern, search frobbase - otherwise, search rulebase
      (cond ((frob-pattern? (term-part goal))
	     (setf frobs (get-frobs (frob-pat-name (term-part goal))))
	     ; match frob pattern against frob database first
	     ; always unify frob patterns at level 0
	     (let ((newenv (search-frobs env goals frobs *top-level* back)))
	       ; if no match against the frob database, try the rulebase
	       (if (impossible? newenv)
		  (search-rules env goals
	            (get-prolog-rules 
		     (expand-term
		      (tag-term (tag-part goal)
				(frob-text-name (frob-pat-name (term-part goal))))
		      env))
		    level back)
		  newenv)))
	    (t
	     (search-rules env goals
	       (get-prolog-rules (functor (expand-term goal env))) level back))))))

(defun search-frobs (env goals rem-frobs level back)
  (let ((goal (term-part (first goals))))
    (if (null rem-frobs)
      (continue-on back)
      (let ((newenv (match-frob env goal (first rem-frobs) level)))
	(if (impossible? newenv)
	  ; unify fails, try other frobs
	  (search-frobs env goals (rest rem-frobs) level back)
	  ; unified, try other goals
	  (pl-search newenv
		     (tag-terms level nil (rest goals))
		     (1+ level)
		     (make-cont :env env :goals goals
				:rules (rest rem-frobs)
				:level level :back back)))))))

;; continue - continue searching with continuation
;; this is used to backtrack to a choice point and continue executing

(defun continue-on (cont)
  (if (null cont)
    *impossible*
    (if (null (cont-goals cont))
      (continue-on (cont-back cont))
      (cond ((frob-pattern? (term-part (first (cont-goals cont))))
	     (search-frobs (cont-env cont) (cont-goals cont) (cont-rules cont)
			   (cont-level cont) (cont-back cont)))
	    (t 
	     (search-rules (cont-env cont) (cont-goals cont) (cont-rules cont)
			   (cont-level cont) (cont-back cont)))))))

;; parse-frob-pattern - parse the argument list of a frob pattern message
;; the argument list looks like 
;;   (pat1 pat2 pat3) equivalent to (:out pat1 pat2 pat3), or
;;   (:in arg1 arg2 arg3 :out pat1 pat2 pat3)
;; the :in arguments are values to be sent to the message 
;; the :out arguments are patterns to be applied to the message's return values

(defun parse-frob-pattern (args)
  (let ((out-seen nil)
	(in-seen nil)
        (msg-args nil)
	(out-pats nil))
    (dolist (arg args)
      (if out-seen
	(push arg out-pats)
        (case arg
	  (:in (setf in-seen t))
	  (:out (setf out-seen t))
	  (otherwise
	    (if in-seen
	      (push arg msg-args)
	      (push arg out-pats))))))
    (values (nreverse msg-args) (nreverse out-pats))))

;; match-frob
;; a pattern is an array [frob (slot1 value1) (slot2 value2 value3) ...]
;; both frob and the values may be logical variables whose bindings will
;; be added to the environment - if failure, return the original environment

(defun match-frob (env pattern frob level)
  (let ((frob-name nil)
	(return-env env))
    (multiple-value-setq (frob-name return-env)
      (parse-frob-name level pattern frob return-env))
    (if frob-name   
      (dolist (query (frob-slots pattern))
        (let ((slot-name (car query))
	      (msg-args (cdr query))
	      (out-pats nil)
	      (ret-vals nil))
          ; get the msg inputs and the output patterns for the pattern
          (multiple-value-setq (msg-args out-pats)
            (parse-frob-pattern msg-args))
          (if (> (length out-pats) 1)
            (progn
	     (format t "Only one method pattern allowed (~s)~%" out-pats)
	     (setf out-pats (list (car out-pats)))))
	  ; call the message, binding ret-vals to its return values
;	  (setf ret-vals
;	    (multiple-value-list (apply slot-name `(,frob ,@msg-args))))
          (setf ret-vals (list (apply slot-name `(,frob ,@msg-args))))
	  ; compare the return values against the output patterns,
          ; adding logical variables to the environment if necessary
	  (if (equal (length out-pats) (length ret-vals))
	    (do ((val ret-vals (cdr val))
		 (pat out-pats (cdr pat)))
		((null val) return-env)
	      ; unify pattern with return value
	      (setf return-env (unify return-env (tag-term level pat)
				                 (tag-term level val)))
	      (if (impossible? return-env)
	        (return (setf return-env env))))
	    (format t "Frob pattern args do not match message return args")))))
    return-env))

(defun parse-frob-name (level pattern frob env)
  (let ((frob-pat (frob-pat-name pattern)))
    (cond ((var-name? frob-pat)
	   (if (equal (frob-class frob-pat) (frob-type frob))
	     (values (frob-type frob)
		     (bind (tag-term level frob-pat)
			   (tag-term level frob) env))
	     (values nil env)))
	  ((or (frob-p frob-pat) (class-frob-p frob-pat))
	   (if (equal frob-pat frob)
	     (values (frob-name frob) env)
	     (values nil env)))
	  ((symbolp frob-pat)
	   (if (equal frob-pat (frob-name frob))
	     (values (frob-name frob) env)
	     (values nil env)))
	  (t (values nil env)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;                            the end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
